#! /usr/bin/env python

"""
Module with various functions.
"""

from __future__ import division
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['dist',
           'frame_center',
           'get_square',
           'get_square_robust',
           'get_circle',
           'get_ellipse',
           'get_annulus_segments',
           'get_annulus',
           'get_annulus_quad',
           'get_annulus_cube',
           'get_ell_annulus',
           'mask_circle',
           'create_ringed_spider_mask',
           'matrix_scaling',
           'prepare_matrix',
           'reshape_matrix']

import numpy as np
from skimage.draw import polygon, circle
from sklearn.preprocessing import scale


def create_ringed_spider_mask(im_shape, ann_out, ann_in=0, sp_width=10, sp_angle=0):
    """
    Mask out information is outside the annulus and inside the spiders (zeros).

    Parameters
    ----------
    im_shape : tuple of int
        Tuple of length two with 2d array shape (Y,X).
    ann_out : int
        Outer radius of the annulus.
    ann_in : int
        Inner radius of the annulus.
    sp_width : int
        Width of the spider arms (3 branches).
    sp_angle : int
        angle of the first spider arm (on the positive horizontal axis) in
        counter-clockwise sense.

    Returns
    -------
    mask : array_like
        2d array of zeros and ones.

    """
    mask = np.zeros(im_shape)

    s = im_shape[0]
    r = s/2.
    theta = np.arctan2(sp_width/2., r)

    t0 = np.array([theta,np.pi-theta,np.pi+theta,np.pi*2.-theta])
    t1 = t0 + sp_angle/180. * np.pi
    t2 = t1 + np.pi/3.
    t3 = t2 + np.pi/3.

    x1 = r * np.cos(t1) + s/2.
    y1 = r * np.sin(t1) + s/2.
    x2 = r * np.cos(t2) + s/2.
    y2 = r * np.sin(t2) + s/2.
    x3 = r * np.cos(t3) + s/2.
    y3 = r * np.sin(t3) + s/2.

    rr1, cc1 = polygon(y1, x1)
    rr2, cc2 = polygon(y2, x2)
    rr3, cc3 = polygon(y3, x3)

    cy, cx = frame_center(mask)
    rr0, cc0 = circle(cy, cx, min(ann_out, cy))
    rr4, cc4 = circle(cy, cx, ann_in)

    mask[rr0, cc0] = 1
    mask[rr1, cc1] = 0
    mask[rr2, cc2] = 0
    mask[rr3, cc3] = 0
    mask[rr4, cc4] = 0
    return mask


def dist(yc,xc,y1,x1):
    """ Returns the Euclidean distance between two points.
    """
    return np.sqrt((yc-y1)**2+(xc-x1)**2)


def frame_center(array, verbose=False):
    """ Returns the coordinates y,x of an image center.
    """   
    cy = array.shape[0]/2. - 0.5
    cx = array.shape[1]/2. - 0.5

    if verbose:
        print('Center px coordinates at x,y = ({:},{:})'.format(cy, cx))
    return cy, cx

    
def get_square(array, size, y, x, position=False, force=False):
    """ Returns an square subframe from a 2d array or image.
    
    Parameters
    ----------
    array : array_like
        Input frame.
    size : int
        Size of the subframe.
    y : int
        Y coordinate of the center of the subframe (obtained with the function
        ``frame_center``).
    x : int
        X coordinate of the center of the subframe (obtained with the function
        ``frame_center``).
    position : bool, optional
        If set to True return also the coordinates of the bottom-left vertex.
    force : bool, optional
        Size and the size of the 2d array must be both even or odd. With
        ``force`` set to True this condition can be avoided.
        
    Returns
    -------
    array_view : array_like
        Sub array.
        
    """
    if not array.ndim == 2:
        raise TypeError('Input array is not a frame or 2d array.')

    ary, arx = array.shape

    if not force:
        if ary % 2 == 0:    # assuming square frames
            if size % 2 != 0:
                size += 1
                print('`Size` is odd (while frame size is even)')
                print('Setting `size` to {} pixels'.format(size))
        else:
            if size % 2 == 0:
                size += 1
                print('`Size` is even (while frame size is odd)')
                print('Setting `size` to {} pixels'.format(size))

    # wing is added to the sides of the subframe center
    if size % 2 != 0:
        wing = int(np.floor(size / 2.))
    else:
        wing = (size / 2.) - 0.5

    y0 = int(y-wing)
    y1 = int(y+wing+1)  # +1 cause endpoint is excluded when slicing
    x0 = int(x-wing)
    x1 = int(x+wing+1)
    if (y0 or x0) < 0 or (y1 or x1) > ary:   # assuming square frames
        msg = 'square cannot be obtained given the size and y,x combination'
        raise RuntimeError(msg)

    array_view = array[y0: y1, x0: x1].copy()
    
    if position:
        return array_view, y-wing, x-wing
    else:
        return array_view


# TODO: Check this function is needed, if not remove and replace by get_square
# TODO: Check handling even sized frames
def get_square_robust(array, size, y, x, position=False, 
                      out_borders='reduced_square', return_wings=False,
                      strict=False):                 
    """ 
    Returns a square subframe from a larger array robustly (different options in
    case the requested subframe outpasses the borders of the larger array.
    
    Parameters
    ----------
    array : array_like
        Input frame.
    size : int, ideally odd
        Size of the subframe to be returned.
    y, x : int
        Coordinates of the center of the subframe.
    position : bool, {False, True}, optional
        If set to True, returns also the coordinates of the left bottom vertex.
    out_borders: string {'reduced_square','rectangular', 'whatever'}, optional
        Option that set what to do if the provided size is such that the 
        sub-array exceeds the borders of the array:
            - 'reduced_square' (default) -> returns a smaller square sub-array: 
            the biggest that fits within the borders of the array (warning msg)
            - 'rectangular' -> returns a cropped sub-array with only the part 
            that fits within the borders of the array; thus a rectangle (warning
            msg)
            - 'whatever' -> returns a square sub-array of the requested size, 
            but filled with zeros where it outpasses the borders of the array 
            (warning msg)
    return_wings: bool, {False,True}, optional
        If True, the function only returns the size of the sub square
        (this can be used to test that there will not be any size reduction of 
        the requested square beforehand)
    strict: bool, {False, True}, optional
        Set to True when you want an error to be raised if the size is not an 
        odd number. Else, the subsquare will be computed even if size is an even
        number. In the later case, the center is placed in such a way that 
        frame_center function of the sub_array would give the input center 
        (at pixel = half dimension minus 1).
        
    Returns
    -------
    default:
    array_view : array_like
        Sub array of the requested dimensions (or smaller depending on its 
        location in the original array and the selected out_borders option)

    if position is set to True and return_wing to False: 
    array_view, y_coord, x_coord: array_like, int, int 
        y_coord and x_coord are the indices of the left bottom vertex

    if return_wing is set to True: 
    wing: int
        the semi-size of the square in agreement with the out_borders option
    """
    if not array.ndim == 2:
        raise TypeError('Input array is not a frame or 2d array.')
    
    n_y = array.shape[0]
    n_x = array.shape[1]

    if strict:
        if size%2==0: 
            raise ValueError('The given size of the sub-square should be odd.')
        wing_bef = int((size-1)/2)
        wing_aft = wing_bef
    else:
        if size%2==0:
            wing_bef = (size/2)-1
            wing_aft = size/2
        else:
            wing_bef = int((size-1)/2)
            wing_aft = wing_bef

    #Consider the case of the sub-array exceeding the array
    if (y-wing_bef < 0 or y+wing_aft+1 >= n_y or x-wing_bef < 0 or 
        x+wing_aft+1 >= n_x):
        if out_borders=='reduced_square':
            wing_bef = min(y,x,n_y-1-y,n_x-1-x)
            wing_aft = wing_bef
            msg = "!!! WARNING: The size of the square sub-array was reduced"+\
                  " to fit within the borders of the array. Now, wings = "+\
                  str(wing_bef)+"px x "+str(wing_aft)+ "px !!!"
            print(msg)
        elif out_borders=='rectangular':
            wing_y = min(y,n_y-1-y)
            wing_x = min(x,n_x-1-x)
            y_init = y-wing_y
            y_fin = y+wing_y+1
            x_init = x-wing_x
            x_fin = x+wing_x+1
            array_view = array[int(y_init):int(y_fin), int(x_init):int(x_fin)]
            msg = "!!! WARNING: The square sub-array was changed to a "+\
                  "rectangular sub-array to fit within the borders of the "+\
                  "array. Now, [y_init,yfin]= ["+ str(y_init)+", "+ str(y_fin)+\
                  "] and [x_init,x_fin] = ["+ str(x_init)+", "+ str(x_fin)+ "]."
            print(msg)
            if position:
                return array_view, y_init, x_init
            else:
                return array_view
        else:
            msg = "!!! WARNING: The square sub-array was not changed but it"+\
                  " exceeds the borders of the array."
            print(msg)

    if return_wings:
        return wing_bef,wing_aft
    
    else:
        # wing is added to the sides of the subframe center. Note the +1 when 
        # closing the interval (python doesn't include the endpoint)
        array_view = array[int(y-wing_bef):int(y+wing_aft+1),
                           int(x-wing_bef):int(x+wing_aft+1)].copy()
        if position:
            return array_view, y-wing_bef, x-wing_bef
        else:
            return array_view


def get_circle(array, radius, output_values=False, cy=None, cx=None):           
    """Returns a centered circular region from a 2d ndarray. All the rest 
    pixels are set to zeros. 
    
    Parameters
    ----------
    array : array_like
        Input 2d array or image. 
    radius : int
        The radius of the circular region.
    output_values : {False, True}
        Sets the type of output.
    cy, cx : int
        Coordinates of the circle center.
        
    Returns
    -------
    values : array_like
        1d array with the values of the pixels in the circular region.
    array_masked : array_like
        Input array with the circular mask applied.
    """
    if not array.ndim == 2:
        raise TypeError('Input array is not a frame or 2d array.')
    sy, sx = array.shape
    if not cy and not cx:
        cy, cx = frame_center(array, verbose=False)
         
    yy, xx = np.ogrid[:sy, :sx]                                                 # ogrid is a multidim mesh creator (faster than mgrid)
    circle = (yy - cy)**2 + (xx - cx)**2                                        # eq of circle. squared distance to the center                                        
    circle_mask = circle < radius**2                                            # mask of 1's and 0's                                       
    if output_values:
        values = array[circle_mask]
        return values
    else:
        array_masked = array*circle_mask
        return array_masked


def get_ellipse(array, a, b, PA, output_values=False, cy=None, cx=None,
                output_indices=False):
    """ Returns a centered elliptical region from a 2d ndarray. All the rest 
    pixels are set to zeros.
    
    Parameters
    ----------
    array : array_like
        Input 2d array or image. 
    a : float or int
        Semi-major axis.
    b : float or int
        Semi-minor axis.
    PA : deg, float
        The PA of the semi-major axis.
    output_values : {False, True}, optional
        If True returns the values of the pixels in the annulus.
    cy, cx : int
        Coordinates of the circle center.
    output_indices : {False, True}, optional
        If True returns the indices inside the annulus.
    
    Returns
    -------
    Depending on output_values, output_indices:
    values : array_like
        1d array with the values of the pixels in the circular region.
    array_masked : array_like
        Input array with the circular mask applied.
    y, x : array_like
        Coordinates of pixels in circle.
    """

    if not array.ndim == 2:
        raise TypeError('Input array is not a frame or 2d array.')
    sy, sx = array.shape
    if not cy and not cx:
        cy, cx = frame_center(array, verbose=False)

    # Definition of other parameters of the ellipse
    f = np.sqrt(a ** 2 - b ** 2)  # distance between center and foci of the ellipse
    PA_rad = np.deg2rad(PA)
    pos_f1 = (cy + f * np.cos(PA_rad), cx + f * np.sin(PA_rad))  # coords of first focus
    pos_f2 = (cy - f * np.cos(PA_rad), cx - f * np.sin(PA_rad))  # coords of second focus

    yy, xx = np.ogrid[:sy, :sx]
    # ogrid is a multidim mesh creator (faster than mgrid)
    ellipse = dist(yy, xx, pos_f1[0], pos_f1[1]) + dist(yy, xx, pos_f2[0],
                                                        pos_f2[1])
    ellipse_mask = ellipse < 2 * a  # mask of 1's and 0's

    if output_values and not output_indices:
        values = array[ellipse_mask]
        return values
    elif output_indices and not output_values:
        indices = np.array(np.where(ellipse_mask))
        y = indices[0]
        x = indices[1]
        return y, x
    elif output_indices and output_values:
        msg = 'output_values and output_indices cannot be both True.'
        raise ValueError(msg)
    else:
        array_masked = array * ellipse_mask
        return array_masked


def get_annulus_segments(array, inner_radius, width, nsegm=8, theta_init=0,
                         output_values=False):
    """ Returns indices or values in segments of a centerered annulus from a
    2d ndarray.

    Parameters
    ----------
    array : array_like
        Input 2d array or image.
    inner_radius : int
        The inner radius of the donut region.
    width : int
        The size of the annulus.
    nsegm : int
        Number of segments of annulus to be extracted.
    theta_init : int
        Initial azimuth [degrees] of the first segment, counting from the postivie
        y-axis clockwise.
    output_values : {False, True}, optional
        If True returns the values of the pixels in the each quadrant instead
        of the indices.

    Returns
    -------
    indices : list with lenght nsegm
        Coordinates of pixels for each annulus segment.
    If output_values is True the pixel values are returned instead.

    """
    if not array.ndim == 2:
        raise TypeError('Input array is not a frame or 2d array.')
    if not isinstance(nsegm, int):
        raise TypeError('nsegm must be an integer.')
    if not isinstance(inner_radius, int):
        raise TypeError('inner_radius must be an integer.')
    if not isinstance(width, int):
        raise TypeError('width must be an integer.')

    cy, cx = frame_center(array)
    azimuth_coverage = np.deg2rad(int(np.ceil(360 / nsegm)))
    twopi = 2 * np.pi

    xx, yy = np.mgrid[:array.shape[0], :array.shape[1]]
    rad = np.sqrt((xx - cx) ** 2 + (yy - cy) ** 2)
    phi = np.arctan2(yy - cy, xx - cx)
    phirot = phi % (twopi)
    outer_radius = inner_radius + width
    indices = []

    for i in range(nsegm):
        phi_start = np.deg2rad(theta_init) + (i * azimuth_coverage)
        phi_end = phi_start + azimuth_coverage

        if phi_start < twopi and phi_end > twopi:
            indices.append(
                np.where((rad >= inner_radius) & (rad < outer_radius) &
                         (phirot >= phi_start) & (phirot <= twopi) |
                         (rad >= inner_radius) & (rad < outer_radius) &
                         (phirot >= 0) & (phirot < phi_end - twopi)))
        elif phi_start >= twopi and phi_end > twopi:
            indices.append(
                np.where((rad >= inner_radius) & (rad < outer_radius) &
                         (phirot >= phi_start - twopi) &
                         (phirot < phi_end - twopi)))
        else:
            indices.append(
                    np.where((rad >= inner_radius) & (rad < outer_radius) &
                             (phirot >= phi_start) & (phirot < phi_end)))

    if output_values:
        values = [array[segment] for segment in indices]
        return np.array(values)
    else:
        return indices


### TODO: remove this in VIP v1.0.0. Replaced with get_annulus_segments
def get_annulus(array, inner_radius, width, output_values=False, 
                output_indices=False):                                          
    """Returns a centerered annulus from a 2d ndarray. All the rest pixels are 
    set to zeros. 
    
    Parameters
    ----------
    array : array_like
        Input 2d array or image. 
    inner_radius : float
        The inner radius of the donut region.
    width : int
        The size of the annulus.
    output_values : {False, True}, optional
        If True returns the values of the pixels in the annulus.
    output_indices : {False, True}, optional
        If True returns the indices inside the annulus.
    
    Returns
    -------
    Depending on output_values, output_indices:
    values : array_like
        1d array with the values of the pixels in the annulus.
    array_masked : array_like
        Input array with the annular mask applied.
    y, x : array_like
        Coordinates of pixels in annulus.
    """
    if not array.ndim == 2:
        raise TypeError('Input array is not a frame or 2d array.')
    array = array.copy()
    cy, cx = frame_center(array)
    yy, xx = np.mgrid[:array.shape[0], :array.shape[1]]
    circle = np.sqrt((xx - cx)**2 + (yy - cy)**2)                                                                               
    donut_mask = (circle <= (inner_radius + width)) & (circle >= inner_radius)
    if output_values and not output_indices:
        values = array[donut_mask]
        return values
    elif output_indices and not output_values:      
        indices = np.array(np.where(donut_mask))
        y = indices[0]
        x = indices[1]
        return y, x
    elif output_indices and output_values:
        raise ValueError('output_values and output_indices cannot be both True.')
    else:
        array_masked = array*donut_mask
        return array_masked
    

### TODO: remove this in VIP v1.0.0. Replaced with get_annulus_segments
def get_annulus_quad(array, inner_radius, width, output_values=False):                                          
    """ Returns indices or values in quadrants of a centerered annulus from a 
    2d ndarray. 
    
    Parameters
    ----------
    array : array_like
        Input 2d array or image. 
    inner_radius : int
        The inner radius of the donut region.
    width : int
        The size of the annulus.
    output_values : {False, True}, optional
        If True returns the values of the pixels in the each quadrant instead
        of the indices.
    
    Returns
    -------
    Depending on output_values:
    values : array_like with shape [4, npix]
        Array with the values of the pixels in each quadrant of annulus.
    ind : array_like with shape [4,2,npix]
        Coordinates of pixels for each quadrant in annulus.
        
    """
    if not array.ndim == 2:
        raise TypeError('Input array is not a frame or 2d array.')
    
    cy, cx = frame_center(array)
    xx, yy = np.mgrid[:array.shape[0], :array.shape[1]]
    circle = np.sqrt((xx - cx)**2 + (yy - cy)**2)                                                                               
    q1 = (circle >= inner_radius) & (circle <= (inner_radius + width)) & (xx >= cx) & (yy <= cy)  
    q2 = (circle >= inner_radius) & (circle <= (inner_radius + width)) & (xx <= cx) & (yy <= cy)
    q3 = (circle >= inner_radius) & (circle <= (inner_radius + width)) & (xx <= cx) & (yy >= cy)
    q4 = (circle >= inner_radius) & (circle <= (inner_radius + width)) & (xx >= cx) & (yy >= cy)
    
    if output_values:
        values = [array[mask] for mask in [q1,q2,q3,q4]]
        return np.array(values)
    else:      
        ind = [np.array(np.where(mask)) for mask in [q1,q2,q3,q4]]          
        return np.array(ind)


### TODO: VIP v1.0.0: make use of get_annulus_segments instead
def get_annulus_cube(array, inner_radius, width, output_values=False):     
    """ Returns a centerered annulus from a 3d ndarray. All the rest pixels are 
    set to zeros. 
    
    Parameters
    ----------
    array : array_like
        Input 2d array or image. 
    inner_radius : int
        The inner radius of the donut region.
    width : int
        The size of the annulus.
    output_values : {False, True}, optional
        If True returns the values of the pixels in the annulus.
    
    Returns
    -------
    Depending on output_values:
    values : array_like
        1d array with the values of the pixels in the circular region.
    array_masked : array_like
        Input array with the annular mask applied.

    """
    if not array.ndim == 3:
        raise TypeError('Input array is not a cube or 3d array.')
    arr_annulus = np.empty_like(array)
    if output_values:
        values = []
        for i in range(array.shape[0]):
            values.append(get_annulus(array[i], inner_radius, width,
                                      output_values=True))
        return np.array(values)
    else:
        for i in range(array.shape[0]):
            arr_annulus[i] = get_annulus(array[i], inner_radius, width)
        return arr_annulus
    

def get_ell_annulus(array, a, b, PA, width, output_values=False,
                    output_indices=False, cy=None, cx=None):
    """Returns a centered elliptical annulus from a 2d ndarray. All the rest 
    pixels are set to zeros. 

    Parameters
    ----------
    array : array_like
        Input 2d array or image. 
    a : flt
        Semi-major axis.
    b : flt
        Semi-minor axis.
    PA : deg
        The PA of the semi-major axis.
    width : flt
        The size of the annulus along the semi-major axis; it is proportionnally 
        thinner along the semi-minor axis).
    output_values : {False, True}, optional
        If True returns the values of the pixels in the annulus.
    output_indices : {False, True}, optional
        If True returns the indices inside the annulus.
    cy,cx: float, optional
        Location of the center of the annulus to be defined. If not provided, 
    it assumes the annuli are centered on the frame.

    Returns
    -------
    Depending on output_values, output_indices:
    values : array_like
        1d array with the values of the pixels in the annulus.
    array_masked : array_like
        Input array with the annular mask applied.
    y, x : array_like
        Coordinates of pixels in annulus.

    """
    if not array.ndim == 2:
        raise TypeError('Input array is not a frame or 2d array.')
    if cy is None or cx is None:
        cy, cx = frame_center(array)
    sy, sx = array.shape

    width_a = width
    width_b = width * b / a

    # Definition of big ellipse
    f_big = np.sqrt((a + width_a / 2.) ** 2 - (
    b + width_b / 2.) ** 2)  # distance between center and foci of the ellipse
    PA_rad = np.deg2rad(PA)
    pos_f1_big = (cy + f_big * np.cos(PA_rad),
                  cx + f_big * np.sin(PA_rad))  # coords of first focus
    pos_f2_big = (cy - f_big * np.cos(PA_rad),
                  cx - f_big * np.sin(PA_rad))  # coords of second focus

    # Definition of small ellipse
    f_sma = np.sqrt((a - width_a / 2.) ** 2 - (
    b - width_b / 2.) ** 2)  # distance between center and foci of the ellipse
    pos_f1_sma = (cy + f_sma * np.cos(PA_rad),
                  cx + f_sma * np.sin(PA_rad))  # coords of first focus
    pos_f2_sma = (cy - f_sma * np.cos(PA_rad),
                  cx - f_sma * np.sin(PA_rad))  # coords of second focus

    yy, xx = np.ogrid[:sy, :sx]
    big_ellipse = dist(yy, xx, pos_f1_big[0], pos_f1_big[1]) + \
                  dist(yy, xx, pos_f2_big[0], pos_f2_big[ 1])
    small_ellipse = dist(yy, xx, pos_f1_sma[0], pos_f1_sma[1]) + \
                    dist(yy, xx, pos_f2_sma[0], pos_f2_sma[1])
    ell_ann_mask = (big_ellipse < 2 * (a + width / 2.)) & (small_ellipse >= 2 * (a - width / 2.))  # mask of 1's and 0's

    if output_values and not output_indices:
        values = array[ell_ann_mask]
        return values
    elif output_indices and not output_values:
        indices = np.array(np.where(ell_ann_mask))
        y = indices[0]
        x = indices[1]
        return y, x
    elif output_indices and output_values:
        msg = 'output_values and output_indices cannot be both True.'
        raise ValueError(msg)
    else:
        array_masked = array * ell_ann_mask
        return array_masked


def mask_circle(array, radius):                                      
    """ Masks (sets pixels to zero) a centered circle from a frame or cube. 
    
    Parameters
    ----------
    array : array_like
        Input frame or cube.
    radius : int
        Radius of the circular aperture.
    
    Returns
    -------
    array_masked : array_like
        Masked frame or cube.
        
    """
    if len(array.shape) == 2:
        sy, sx = array.shape
        cy, cx = frame_center(array)
        xx, yy = np.ogrid[:sy, :sx]
        circle = (xx - cx)**2 + (yy - cy)**2    # squared distance to the center
        hole_mask = circle > radius**2                                             
        array_masked = array*hole_mask
        
    if len(array.shape) == 3:
        n, sy, sx = array.shape
        cy, cx = frame_center(array[0])
        xx, yy = np.ogrid[:sy, :sx]
        circle = (xx - cx)**2 + (yy - cy)**2    # squared distance to the center
        hole_mask = circle > radius**2      
        array_masked = np.empty_like(array)
        for i in range(n):
            array_masked[i] = array[i]*hole_mask
        
    return array_masked


def matrix_scaling(matrix, scaling):
    """ Scales a matrix using sklearn.preprocessing.scale function.

    scaling : {None, 'temp-mean', 'spat-mean', 'temp-standard', 'spat-standard'}
    With None, no scaling is performed on the input data before SVD. With
    "temp-mean" then temporal px-wise mean subtraction is done, with
    "spat-mean" then the spatial mean is subtracted, with "temp-standard"
    temporal mean centering plus scaling to unit variance is done and with
    "spat-standard" spatial mean centering plus scaling to unit variance is
    performed.
    """
    if scaling is None:
        pass
    elif scaling == 'temp-mean':
        matrix = scale(matrix, with_mean=True, with_std=False)
    elif scaling == 'spat-mean':
        matrix = scale(matrix, with_mean=True, with_std=False, axis=1)
    elif scaling == 'temp-standard':
        matrix = scale(matrix, with_mean=True, with_std=True)
    elif scaling == 'spat-standard':
        matrix = scale(matrix, with_mean=True, with_std=True, axis=1)
    else:
        raise ValueError('Scaling mode not recognized')

    return matrix


def prepare_matrix(array, scaling=None, mask_center_px=None, mode='fullfr',
                   annulus_radius=None, annulus_width=None, verbose=True):
    """ Builds the matrix for the SVD/PCA and other matrix decompositions,
    centers the data and masks the frames central area if needed.

    Parameters
    ----------
    array : array_like
        Input cube, 3d array.
    scaling : {None, 'temp-mean', 'spat-mean', 'temp-standard', 'spat-standard'}
        With None, no scaling is performed on the input data before SVD. With
        "temp-mean" then temporal px-wise mean subtraction is done, with
        "spat-mean" then the spatial mean is subtracted, with "temp-standard"
        temporal mean centering plus scaling to unit variance is done and with
        "spat-standard" spatial mean centering plus scaling to unit variance is
        performed.
    mask_center_px : None or Int, optional
        Whether to mask the center of the frames or not.
    mode : {'fullfr', 'annular'}
        Whether to use the whole frames or a single annulus.
    annulus_radius : float
        Distance in pixels from the center of the frame to the center of the
        annulus.
    annulus_width : float
        Width of the annulus in pixels.
    verbose : {True, False}, bool optional
        If True prints intermediate info and timing.

    Returns
    -------
    If mode is `annular` then the indices of the annulus (yy, xx) are returned
    along with the matrix.

    matrix : array_like
        Out matrix whose rows are vectorized frames from the input cube.

    """
    if mode == 'annular':
        if annulus_radius is None or annulus_width is None:
            msgerr = 'Annulus_radius and/or annulus_width can be None in annular '
            msgerr += 'mode'
            raise ValueError(msgerr)

        ind = get_annulus(array[0], annulus_radius - annulus_width / 2.,
                          annulus_width, output_indices=True)
        yy, xx = ind
        matrix = array[:, yy, xx]

        matrix = matrix_scaling(matrix, scaling)

        if verbose:
            msg = 'Done vectorizing the cube annulus. Matrix shape [{:},{:}]'
            print(msg.format(matrix.shape[0], matrix.shape[1]))
        return matrix, ind

    elif mode == 'fullfr':
        if mask_center_px:
            array = mask_circle(array, mask_center_px)

        nfr = array.shape[0]
        matrix = np.reshape(array, (nfr, -1))  # == for i: array[i].flatten()

        matrix = matrix_scaling(matrix, scaling)

        if verbose:
            msg = 'Done vectorizing the frames. Matrix shape [{:},{:}]'
            print(msg.format(matrix.shape[0], matrix.shape[1]))
        return matrix


def reshape_matrix(array, y, x):
    """ Converts a matrix whose rows are vectorized frames to a cube with
    reshaped frames.
    """
    return array.reshape(array.shape[0], y, x)

        




#! /usr/bin/env python

"""
2d fitting.
"""
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['fit_2dgaussian',
           'fit_2dmoffat']

import numpy as np
import pandas as pd
import photutils
from scipy.optimize import leastsq
from astropy.modeling import models
from astropy.modeling.fitting import LevMarLSQFitter
from astropy.stats import gaussian_sigma_to_fwhm, gaussian_fwhm_to_sigma, sigma_clipped_stats
from .shapes import get_square, frame_center
from .utils_var import pp_subplots


def fit_2dgaussian(array, crop=False, cent=None, cropsize=15, fwhmx=4, fwhmy=4, 
                   theta=0, threshold=False, sigfactor=6, full_output=False, 
                   debug=False):
    """ Fitting a 2D Gaussian to the 2D distribution of the data.
    
    Parameters
    ----------
    array : array_like
        Input frame with a single PSF.
    crop : bool, optional
        If True an square sub image will be cropped.
    cent : tuple of int, optional
        X,Y integer position of source in the array for extracting the subimage. 
        If None the center of the frame is used for cropping the subframe (the 
        PSF is assumed to be ~ at the center of the frame). 
    cropsize : int, optional
        Size of the subimage.
    fwhmx, fwhmy : float, optional
        Initial values for the standard deviation of the fitted Gaussian, in px.
    theta : float, optional
        Angle of inclination of the 2d Gaussian counting from the positive X
        axis.
    threshold : bool, optional
        If True the background pixels (estimated using sigma clipped statistics)
        will be replaced by small random Gaussian noise.
    sigfactor : int, optional
        The background pixels will be thresholded before fitting a 2d Gaussian
        to the data using sigma clipped statistics. All values smaller than
        (MEDIAN + sigfactor*STDDEV) will be replaced by small random Gaussian 
        noise. 
    full_output : bool, optional
        If False it returns just the centroid, if True also returns the 
        FWHM in X and Y (in pixels), the amplitude and the rotation angle.
    debug : bool, optional
        If True, the function prints out parameters of the fit and plots the
        data, model and residuals.
        
    Returns
    -------
    mean_y : float
        Source centroid y position on input array from fitting. 
    mean_x : float
        Source centroid x position on input array from fitting.
        
    If ``full_output`` is True it returns a Pandas dataframe containing the
    following columns:
    'amplitude' : Float value. Amplitude of the Gaussian.
    'centroid_x' : Float value. X coordinate of the centroid.
    'centroid_y' : Float value. Y coordinate of the centroid.
    'fwhm_x' : Float value. FHWM in X [px].
    'fwhm_y' : Float value. FHWM in Y [px].
    'theta' : Float value. Rotation angle.
    
    """
    if not array.ndim == 2:
        raise TypeError('Input array is not a frame or 2d array')
    
    if crop:
        if cent is None:
            ceny, cenx = frame_center(array)
        else:
            cenx, ceny = cent
        
        imside = array.shape[0]
        psf_subimage, suby, subx = get_square(array, min(cropsize, imside), 
                                              ceny, cenx, position=True)  
    else:
        psf_subimage = array.copy()  
    
    if threshold:
        _, clipmed, clipstd = sigma_clipped_stats(psf_subimage, sigma=2)
        indi = np.where(psf_subimage <= clipmed + sigfactor * clipstd)
        subimnoise = np.random.randn(psf_subimage.shape[0],
                                     psf_subimage.shape[1]) * clipstd
        psf_subimage[indi] = subimnoise[indi]

    # Creating the 2D Gaussian model
    init_amplitude = np.ptp(psf_subimage)
    xcom, ycom = photutils.centroid_com(psf_subimage)
    gauss = models.Gaussian2D(amplitude=init_amplitude, theta=theta,
                              x_mean=xcom, y_mean=ycom,
                              x_stddev=fwhmx * gaussian_fwhm_to_sigma,
                              y_stddev=fwhmy * gaussian_fwhm_to_sigma)
    # Levenberg-Marquardt algorithm
    fitter = LevMarLSQFitter()                  
    y, x = np.indices(psf_subimage.shape)
    fit = fitter(gauss, x, y, psf_subimage)

    if crop:
        mean_y = fit.y_mean.value + suby
        mean_x = fit.x_mean.value + subx
    else:
        mean_y = fit.y_mean.value
        mean_x = fit.x_mean.value 
    fwhm_y = fit.y_stddev.value*gaussian_sigma_to_fwhm
    fwhm_x = fit.x_stddev.value*gaussian_sigma_to_fwhm 
    amplitude = fit.amplitude.value
    theta = fit.theta.value
    
    if debug:
        if threshold:  msg = 'Subimage thresholded / Model / Residuals'
        else: msg = 'Subimage (no threshold) / Model / Residuals'
        pp_subplots(psf_subimage, fit(x, y), psf_subimage-fit(x, y), 
                    colorb=True, grid=True, title=msg)
        print('FWHM_y =', fwhm_y)
        print('FWHM_x =', fwhm_x)
        print()
        print('centroid y =', mean_y)
        print('centroid x =', mean_x)
        print('centroid y subim =', fit.y_mean.value)
        print('centroid x subim =', fit.x_mean.value)
        print() 
        print('peak =', amplitude)
        print('theta =', theta)
    
    if full_output:
        return pd.DataFrame({'centroid_y': mean_y, 'centroid_x': mean_x,
                             'fwhm_y': fwhm_y, 'fwhm_x': fwhm_x,
                             'amplitude': amplitude, 'theta': theta}, index=[0])
    else:
        return mean_y, mean_x


    
def fit_2dmoffat(array, yy, xx, full_output=False):
    """Fits a star/planet with a 2D circular Moffat PSF.
    
    Parameters
    ----------
    array : array_like
        Subimage with a single point source, approximately at the center. 
    yy : int
        Y integer position of the first pixel (0,0) of the subimage in the 
        whole image.
    xx : int
        X integer position of the first pixel (0,0) of the subimage in the 
        whole image.
    full_output: bool, opt
        Whether to return floor, height, mean_y, mean_x, fwhm, beta, or just 
        mean_y, mean_x
    
    Returns
    -------
    floor : float
        Level of the sky background (fit result).
    height : float
        PSF amplitude (fit result).
    mean_x : float
        Source centroid x position on the full image from fitting.
    mean_y : float
        Source centroid y position on the full image from fitting. 
    fwhm : float
        Gaussian PSF full width half maximum from fitting (in pixels).
    beta : float
        "beta" parameter of the moffat function.
    """
    def moffat(floor, height, mean_y, mean_x, fwhm, beta): # def Moffat function
        alpha = 0.5*fwhm/np.sqrt(2.**(1./beta)-1.)
        return lambda y,x: floor + height/((1.+(((x-mean_x)**2+(y-mean_y)**2)/\
                                                alpha**2.))**beta)

    def err(p,data):
        return np.ravel(moffat(*p)(*np.indices(data.shape))-data)
    ############################################################################

    maxi = array.max()  # find starting values
    floor = np.ma.median(array.flatten())
    height = maxi - floor
    if height == 0.0:   # if star is saturated it could be that
        floor = np.mean(array.flatten())
        height = maxi - floor

    mean_y = (np.shape(array)[0]-1)/2
    mean_x = (np.shape(array)[1]-1)/2

    fwhm = np.sqrt(np.sum((array > floor+height/2.).flatten()))

    beta = 4
    
    p0 = floor, height, mean_y, mean_x, fwhm, beta

    p = leastsq(err, p0, args=(array), maxfev=1000)
    p = p[0]
    
    # results
    floor = p[0]                                
    height = p[1]
    mean_y = p[2] + yy
    mean_x = p[3] + xx
    fwhm = np.abs(p[4])
    beta = p[5]
    
    if full_output:
        return floor, height, mean_y, mean_x, fwhm, beta
    else:
        return mean_y, mean_x




#!/usr/bin/env python
 
import os
import re
from setuptools import setup
from pip.req import parse_requirements
from setuptools.command.install import install


# Hackishly override of the install method
class InstallReqs(install):
    def run(self):
        print(" ********************** ")
        print(" *** Installing VIP *** ")
        print(" ********************** ")
        os.system('pip install -r requirements')
        install.run(self)


PACKAGE_PATH = os.path.abspath(os.path.join(__file__, os.pardir))

# parse_requirements() returns generator of pip.req.InstallRequirement objects
install_reqs = parse_requirements(os.path.join(PACKAGE_PATH, 'requirements'),
                                  session=False)
requirements = [str(ir.req) for ir in install_reqs]

with open(os.path.join(PACKAGE_PATH, 'readme.rst')) as readme_file:
    README = readme_file.read()

with open(os.path.join(PACKAGE_PATH, 'vip_hci/__init__.py')) as version_file:
    version_file = version_file.read()
    VERSION = re.search(r"""^__version__ = ['"]([^'"]*)['"]""",
                        version_file, re.M)
    VERSION = VERSION.group(1)


PACKAGES = ['vip_hci',
            'vip_hci.conf',
            'vip_hci.exlib',
            'vip_hci.fits',
            'vip_hci.frdiff',
            'vip_hci.leastsq',
            'vip_hci.llsg',
            'vip_hci.madi',
            'vip_hci.negfc',
            'vip_hci.nmf',
            'vip_hci.pca',
            'vip_hci.phot',
            'vip_hci.preproc',
            'vip_hci.stats',
            'vip_hci.var']

setup(
    name='vip_hci',
    version=VERSION,
    description='Package for astronomical high-contrast image processing.',
    long_description=README,
    license='MIT',
    author='Carlos Alberto Gomez Gonzalez',
    author_email='carlosgg33@gmail.com',
    url='https://github.com/vortex-exoplanet/VIP',
    cmdclass={'install': InstallReqs},
    packages=PACKAGES,
    install_requires=requirements,
    zip_safe=False,
    classifiers=['Intended Audience :: Science/Research',
                 'License :: OSI Approved :: MIT License',
                 'Operating System :: MacOS :: MacOS X',
                 'Operating System :: POSIX :: Linux',
                 'Natural Language :: English',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3.6',
                 'Topic :: Scientific/Engineering :: Astronomy'
                 ] 
)


#! /usr/bin/env python

"""
Module with frame/cube filtering functionalities
"""

from __future__ import division

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['frame_filter_highpass',
           'frame_filter_lowpass',
           'cube_filter_highpass',
           'cube_filter_iuwt',
           'frame_filter_gaussian2d',
           'gaussian_kernel']

import warnings
try:
    import cv2
    no_opencv = False
except ImportError:
    msg = "Opencv python bindings are missing."
    warnings.warn(msg, ImportWarning)
    no_opencv = True
import numpy as np
import photutils
import pyprind
from scipy.ndimage import gaussian_filter, median_filter      
from astropy.convolution import convolve_fft, Gaussian2DKernel
from astropy.stats import gaussian_fwhm_to_sigma
from .shapes import frame_center
from ..exlib import iuwt


def cube_filter_iuwt(cube, coeff=5, rel_coeff=1, full_output=False):
    """
    
    Parameters
    ----------
    cube : array_like
        Input cube.
    coeff : int, optional
        Number of wavelet scales to be used in the decomposition.
    rel_coeff : int, optional
        Number of relevant coefficients. In other words how many wavelet scales
        will represent in a better way our data. One or two scales are enough
        for filtering our images.
    full_output : {False, True}, bool optional
        If True, an additional cube with the multiscale decomposition of each
        frame will be returned.
    
    Returns
    -------
    cubeout : array_like
        Output cube with the filtered frames.
    
    If full_output is True the filtered cube is returned together with the a 
    4d cube containing the multiscale decomposition of each frame.
    
    """
    cubeout = np.zeros_like(cube)
    cube_coeff = np.zeros((cube.shape[0], coeff, cube.shape[1], cube.shape[2]))
    n_frames = cube.shape[0]
    
    msg = 'Decomposing frames with the Isotropic Undecimated Wavelet Transform.'
    bar = pyprind.ProgBar(n_frames, stream=1, title=msg)
    for i in range(n_frames):
        res = iuwt.iuwt_decomposition(cube[i], coeff, store_smoothed=False)
        cube_coeff[i] = res
        for j in range(rel_coeff):
            cubeout[i] += cube_coeff[i][j] 
        bar.update()
        
    if full_output:
        return cubeout, cube_coeff
    else:
        return cubeout


def cube_filter_highpass(array, mode='laplacian', median_size=5, kernel_size=5,
                         fwhm_size=5, btw_cutoff=0.2, btw_order=2,
                         verbose=True):
    """ Wrapper of ``frame_filter_highpass`` for cubes or 3d arrays.

    Parameters
    ----------
    array : array_like
        Input 3d array.
    mode : str, optional
        See the documentation of the ``frame_filter_highpass`` function.
    median_size : int, optional
        See the documentation of the ``frame_filter_highpass`` function.
    kernel_size : int, optional
        See the documentation of the ``frame_filter_highpass`` function.
    fwhm_size : int, optional
        See the documentation of the ``frame_filter_highpass`` function.
    btw_cutoff : float
        See the documentation of the ``frame_filter_highpass`` function.
    btw_order : int
        See the documentation of the ``frame_filter_highpass`` function.
    verbose : boolean, optional
        If True timing and progress bar are shown.
    
    Returns
    -------
    filtered : array_like
        High-pass filtered cube.
    """
    if not array.ndim == 3:
        raise TypeError('Input array is not a cube or 3d array')
    
    n_frames = array.shape[0]
    array_out = np.zeros_like(array)
    if verbose:
        msg = 'Applying the high-pass filter on cube frames:'
        bar = pyprind.ProgBar(n_frames, stream=1, title=msg, bar_char='.')
    for i in range(n_frames):
        array_out[i] = frame_filter_highpass(array[i], mode, median_size, 
                                            kernel_size, fwhm_size, btw_cutoff, 
                                            btw_order)
        if verbose:
            bar.update()
        
    return array_out
    

def fft(array):
    """ Performs the 2d discrete Fourier transform (using numpy's fft2 function) 
    on the data from the original image. This produces a new representation of 
    the image in which each pixel represents a spatial frequency and 
    orientation, rather than an xy coordinate. When Fourier-transformed images 
    are plotted graphically, the low frequencies are found at the centre; this 
    is not what fft2 actually produces, so we need to also apply numpy's 
    fftshift (centering low frequencies).
    """
    fft_array = np.fft.fftshift(np.fft.fft2(array))
    return fft_array


def ifft(array):
    """ Gets the inverse Fourier transform on the image. This produces an array 
    of complex numbers whose absolute values correspond to the image in the 
    original space (decentering).
    """
    new_array = np.abs(np.fft.ifft2(np.fft.ifftshift(array)))
    return new_array


def frame_filter_highpass(array, mode, median_size=5, kernel_size=5, 
                          fwhm_size=5, btw_cutoff=0.2, btw_order=2):
    """ High-pass filtering of input frame depending on parameter ``mode``. The
    filtered image properties will depend on the ``mode`` and the relevant
    parameters.
    
    Parameters
    ----------
    array : array_like
        Input array, 2d frame.
    mode : {'laplacian', 'laplacian-conv', 'median-subt', 'gauss-subt', 'fourier-butter'}
        Type of High-pass filtering. ``laplacian`` applies a Laplacian fiter
        with kernel size defined by ``kernel_size`` using the Opencv library.
        ``laplacian-conv`` applies a Laplacian high-pass filter by defining a
        kernel (with ``kernel_size``) and using the ``convolve_fft`` Astropy
        function. ``median-subt`` subtracts a median low-pass filtered version
        of the image. ``gauss-subt`` subtracts a Gaussian low-pass filtered
        version of the image. ``fourier-butter`` applies a high-pass 2D
        Butterworth filter in Fourier domain.
    median_size : int, optional
        Size of the median box for filtering the low-pass median filter.
    kernel_size : int, optional
        Size of the Laplacian kernel used in ``laplacian`` mode. It must be an
        positive odd integer value.
    fwhm_size : int, optional
        Size of the Gaussian kernel used in ``gaus-subt`` mode.
    btw_cutoff : float
        Frequency cutoff for low-pass 2d Butterworth filter used in
        ``fourier-butter`` mode.
    btw_order : int
        Order of low-pass 2d Butterworth filter used in ``fourier-butter`` mode.
    
    Returns
    -------
    filtered : array_like
        High-pass filtered image.
        
    """
    def butter2d_lp(size, cutoff, n=3):
        """ Create low-pass 2D Butterworth filter. 
        Function from PsychoPy library, credits to Jonathan Peirce, 2010

        Parameters
        ----------
        size : tuple
            size of the filter
        cutoff : float
            relative cutoff frequency of the filter (0 - 1.0)
        n : int, optional
            order of the filter, the higher n is the sharper
            the transition is.
        
        Returns
        -------
        numpy.ndarray
          filter kernel in 2D centered
        """
        if not 0 < cutoff <= 1.0:
            raise ValueError('Cutoff frequency must be between 0 and 1.0.')
    
        if not isinstance(n, int):
            raise ValueError('n must be an integer >= 1.')
    
        rows, cols = size
        x = np.linspace(-0.5, 0.5, cols) * cols
        y = np.linspace(-0.5, 0.5, rows) * rows
    
        # An array with every pixel = radius relative to center
        radius = np.sqrt((x**2)[np.newaxis] + (y**2)[:, np.newaxis])
        # The filter
        f = 1 / (1.0 + (radius / cutoff)**(2*n))   
        return f
    
    #---------------------------------------------------------------------------
    
    if not array.ndim == 2:
        raise TypeError("Input array is not a frame or 2d array.")
    
    if mode == 'laplacian':
        # Applying a Laplacian high-pass kernel
        if kernel_size % 2 == 0 or kernel_size < 0:
            raise ValueError("Kernel size must be an odd and positive value.")
        if not no_opencv:
            msg = "Opencv bindings are missing. Trying a convolution with a "
            msg += "Laplacian kernel instead."

        filtered = cv2.Laplacian(-array, cv2.CV_32F, ksize=kernel_size)

    elif mode == 'laplacian-conv':
        # Applying a Laplacian high-pass kernel defining a kernel and using
        # the convolve_fft Astropy function
        kernel3 = np.array([[-1, -1, -1],
                            [-1, 8, -1],
                            [-1, -1, -1]])
        kernel5 = np.array([[-4, -1, 0, -1, -4],
                            [-1, 2, 3, 2, -1],
                            [0, 3, 4, 3, 0],
                            [-1, 2, 3, 2, -1],
                            [-4, -1, 0, -1, -4]])
        kernel7 = np.array([[-10, -5, -2, -1, -2, -5, -10],
                            [-5, 0, 3, 4, 3, 0, -5],
                            [-2, 3, 6, 7, 6, 3, -2],
                            [-1, 4, 7, 8, 7, 4, -1],
                            [-2, 3, 6, 7, 6, 3, -2],
                            [-5, 0, 3, 4, 3, 0, -5],
                            [-10, -5, -2, -1, -2, -5, -10]])
        if kernel_size == 3:
            kernel = kernel3
        elif kernel_size == 5:
            kernel = kernel5
        elif kernel_size == 7:
            kernel = kernel7
        else:
            raise ValueError('Kernel size must be either 3, 5 or 7.')
        filtered = convolve_fft(array, kernel, normalize_kernel=False,
                                nan_treatment='fill')
    
    elif mode == 'median-subt':
        # Subtracting the low_pass filtered (median) image from the image itself  
        medianed = median_filter(array, median_size, mode='nearest')
        filtered = array - medianed
    
    elif mode == 'gauss-subt':
        # Subtracting the low_pass filtered (median) image from the image itself  
        gaussed = frame_filter_gaussian2d(array, fwhm_size, mode='conv')
        filtered = array - gaussed
        
    elif mode == 'fourier-butter':
        # Designs an n-th order high-pass 2D Butterworth filter
        filt = butter2d_lp(array.shape, cutoff=btw_cutoff, n=btw_order)
        filt = 1. - filt                        
        array_fft = fft(array)
        fft_new = array_fft * filt
        filtered = ifft(fft_new)        
        
    else:
        raise TypeError('Mode not recognized.')
        
    return filtered


def frame_filter_lowpass(array, mode, median_size=5, fwhm_size=5):
    """ Low-pass filtering of input frame depending on parameter *mode*. 
    
    Parameters
    ----------
    array : array_like
        Input array, 2d frame.
    mode : {'median', 'gauss'}
        Type of low-pass filtering.
    median_size : int
        Size of the median box for filtering the low-pass median filter.
    fwhm_size : int
        Size of the Gaussian kernel for the low-pass Gaussian filter.
    
    Returns
    -------
    filtered : array_like
        Low-pass filtered image.
        
    """
    if not array.ndim==2:
        raise TypeError('Input array is not a frame or 2d array.')
       
    if mode=='median':
        # creating the low_pass filtered (median) image
        filtered = median_filter(array, int(median_size), mode='nearest')
    
    elif mode=='gauss':
        # creating the low_pass filtered (median) image 
        filtered = frame_filter_gaussian2d(array, fwhm_size, mode='conv')    
        
    else:
        raise TypeError('Mode not recognized.')
        
    return filtered


def frame_filter_gaussian2d(array, size_fwhm, mode='conv'):
    """ 2d Gaussian filter. 
    
    Parameters
    ----------
    array : array_like
        Input array, 2d frame.
    size_fwhm : float
        Size in pixels of the FWHM of the gaussian kernel.
    mode : {'conv', 'convfft'}
        'conv' uses the multidimensional gaussian filter from scipy.ndimage and
        'convfft' uses the fft convolution with a 2d Gaussian kernel.
        
    Returns
    -------
    filtered : array_like
        Convolved image.
        
    """
    if not array.ndim==2:
        raise TypeError('Input array is not a frame or 2d array.')
    
    if mode=='conv':
        filtered = gaussian_filter(array, sigma=size_fwhm*gaussian_fwhm_to_sigma, 
                                   order=0, mode='nearest')
    elif mode=='convfft':
        # FFT Convolution with a 2d gaussian kernel created with Astropy.
        gaus = Gaussian2DKernel(stddev=size_fwhm*gaussian_fwhm_to_sigma)
        filtered = convolve_fft(array, gaus)
    else:
        raise TypeError('Mode not recognized.')
    
    return filtered
   
    
def gaussian_kernel(size, size_y=None):
    """ Gaussian kernel.
    """
    size = int(size)
    if not size_y:
        size_y = size
    else:
        size_y = int(size_y)
    x, y = np.mgrid[-size:size+1, -size_y:size_y+1]
    g = np.exp(-(x**2/float(size)+y**2/float(size_y)))

    fwhm = size
    fwhm_aper = photutils.CircularAperture((frame_center(g)), fwhm/2.)
    fwhm_aper_phot = photutils.aperture_photometry(g, fwhm_aper)
    g_norm = g/np.array(fwhm_aper_phot['aperture_sum'])
     
    return g_norm/g_norm.max()






#! /usr/bin/env python

"""
Distance between images.
"""

from __future__ import division

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['cube_distance',
           'cube_distance_to_frame']

import numpy as np
import scipy.stats
from matplotlib import pyplot as plt
from ..var import get_annulus
from skimage.measure import compare_ssim as ssim


def cube_distance(array, frame, mode='full', dist='sad', inradius=None,
                  width=None, plot=True):
    """ Computes the distance (or similarity) between frames in a cube, using
    one as the reference. Depending on the mode, the whole image can be used,
    or just the pixels in a given annulus. The criteria used are:
    - the Manhattan distance (SAD or sum of absolute differences),
    - the Euclidean distance (square root of the sum of the squared differences),
    - the Mean Squared Error,
    - the Spearman correlation coefficient,
    - the Pearson correlation coefficient,
    - the Structural Similarity Index (SSIM).

    The SAD, MSE and Ecuclidean criteria are dissimilarity criteria, which
    means that 0 is perfect similarity.
    The Spearman and Pearson correlation coefficients, vary between -1 and +1
    with 0 implying no correlation. Correlations of -1 or +1 imply an exact
    linear relationship.
    The Structural Similarity Index was proposed by Wang et al. 2004.
    (http://www.cns.nyu.edu/pub/eero/wang03-reprint.pdf)
    SSIM varies between -1 and 1, where 1 means perfect similarity. SSIM
    attempts to model the perceived change in the structural information of the
    image. The mean SSIM is reported.

    Parameters
    ----------
    array : array_like
        Input cube or 3d array.
    frame : int
        Reference frame in the cube.
    mode : {'full','annulus'}, string optional
        Whether to use the full frames or a centered annulus.
    dist : {'sad','euclidean','mse','pearson','spearman', 'ssim'}, str optional
        Which criterion to use.
    inradius : None or int, optional
        The inner radius when mode is 'annulus'.
    width : None or int, optional
        The width when mode is 'annulus'.
    plot : {True, False}, bool optional
        Whether to plot the distances or not.

    Returns
    -------
    lista : array_like
        1d array of distances for each frame wrt the reference one.

    """
    if not array.ndim ==3:
        raise TypeError('The input array is not a cube or 3d array')
    lista = []
    n = array.shape[0]
    if mode=='full':
        frame_ref = array[frame]
    elif mode=='annulus' and inradius and width:
        frame_ref = get_annulus(array[frame], inradius, width, True)
    else:
        raise TypeError('Mode not recognized or missing parameters')

    for i in range(n):
        if mode=='full':
            framei = array[i]
        elif mode=='annulus':
            framei = get_annulus(array[i], inradius, width, True)

        if dist=='sad':
            lista.append(np.sum(abs(frame_ref - framei)))
        elif dist=='euclidean':
            lista.append(np.sqrt(np.sum((frame_ref - framei)**2)))
        elif dist=='mse':
            lista.append((np.sum((frame_ref - framei)**2))/len(frame_ref))
        elif dist=='pearson':
            pears, _ = scipy.stats.pearsonr(frame_ref.ravel(), framei.ravel())
            lista.append(pears)
        elif dist=='spearman':
            spear, _ = scipy.stats.spearmanr(frame_ref.ravel(), framei.ravel())
            lista.append(spear)
        elif dist=='ssim':
            mean_ssim = ssim(frame_ref, framei,
                  win_size=7,
                  dynamic_range=frame_ref.max() - frame_ref.min(),
                  gaussian_weights=True,
                  sigma=1.5,
                  use_sample_covariance=True)
            lista.append(mean_ssim)
        else:
            raise ValueError('Distance not recognized')
    lista = np.array(lista)

    median_cor = np.median(lista)
    mean_cor = np.mean(lista)
    if plot:
        _, ax = plt.subplots(figsize=(12,6))
        x = range(len(lista))
        ax.plot(x, lista, '-', color='blue', alpha=0.3)
        ax.plot(x, lista, '.', color='blue', alpha=0.5)
        ax.vlines(frame, ymin=np.nanmin(lista), ymax=np.nanmax(lista),
                   colors='green', linestyles='dashed', lw=2, alpha=0.8,
                   label='Frame '+str(frame))
        ax.hlines(median_cor, xmin=-1, xmax=n+1, colors='purple',
                   linestyles='solid', label='Median value : '+str(median_cor))
        ax.hlines(mean_cor, xmin=-1, xmax=n+1, colors='red',
                   linestyles='solid', label='Mean value : '+str(mean_cor))
        plt.xlabel('Frame number')
        if dist=='sad':
            plt.ylabel('SAD - Manhattan distance')
        elif dist=='euclidean':
            plt.ylabel('Euclidean distance')
        elif dist=='pearson':
            plt.ylabel('Pearson correlation coefficient')
        elif dist=='spearman':
            plt.ylabel('Spearman correlation coefficient')
        elif dist=='mse':
            plt.ylabel('Mean squared error')
        elif dist=='ssim':
            plt.ylabel('Structural Similarity Index')

        plt.xlim(xmin=-1, xmax=n+1)
        plt.minorticks_on()
        plt.legend(fancybox=True, framealpha=0.5, fontsize=12, loc='best')
        plt.grid(which='both')

    return lista


def cube_distance_to_frame(array, frame_ref, mode='full', dist='sad',
                           inradius=None, width=None, plot=True):
    """ Computes the distance (or similarity) between frames in a cube and a
    reference image. Depending on the mode, the whole image can be used,
    or just the pixels in a given annulus. The criteria used are:
    - the Manhattan distance (SAD or sum of absolute differences),
    - the Euclidean distance (square root of the sum of the squared differences),
    - the Mean Squared Error,
    - the Spearman correlation coefficient,
    - the Pearson correlation coefficient,
    - the Structural Similarity Index (SSIM).

    The SAD, MSE and Ecuclidean criteria are dissimilarity criteria, which
    means that 0 is perfect similarity.
    The Spearman and Pearson correlation coefficients, vary between -1 and +1
    with 0 implying no correlation. Correlations of -1 or +1 imply an exact
    linear relationship.
    The Structural Similarity Index was proposed by Wang et al. 2004.
    (http://www.cns.nyu.edu/pub/eero/wang03-reprint.pdf)
    SSIM varies between -1 and 1, where 1 means perfect similarity. SSIM
    attempts to model the perceived change in the structural information of the
    image. The mean SSIM is reported.

    Parameters
    ----------
    array : array_like
        Input cube or 3d array.
    frame_ref : array_like
        Reference image.
    mode : {'full','annulus'}, string optional
        Whether to use the full frames or a centered annulus.
    dist : {'sad','euclidean','mse','pearson','spearman','ssim'}, str optional
        Which criterion to use.
    inradius : None or int, optional
        The inner radius when mode is 'annulus'.
    width : None or int, optional
        The width when mode is 'annulus'.
    plot : {True, False}, bool optional
        Whether to plot the distances or not.

    Returns
    -------
    lista : array_like
        1d array of distances for each frame wrt the reference one.

    """
    if not array.ndim ==3:
        raise TypeError('The input array is not a cube or 3d array')
    lista = []
    n = array.shape[0]
    if mode=='full':
        frame_ref = frame_ref
    elif mode=='annulus' and inradius and width:
        frame_ref = get_annulus(frame_ref, inradius, width, True)
    else:
        raise TypeError('Mode not recognized or missing parameters')

    for i in range(n):
        if mode=='full':
            framei = array[i]
        elif mode=='annulus':
            framei = get_annulus(array[i], inradius, width, True)

        if dist=='sad':
            lista.append(np.sum(abs(frame_ref - framei)))
        elif dist=='euclidean':
            lista.append(np.sqrt(np.sum((frame_ref - framei)**2)))
        elif dist=='mse':
            lista.append((np.sum((frame_ref - framei)**2))/len(frame_ref))
        elif dist=='pearson':
            pears, _ = scipy.stats.pearsonr(frame_ref.ravel(), framei.ravel())
            lista.append(pears)
        elif dist=='spearman':
            spear, _ = scipy.stats.spearmanr(frame_ref.ravel(), framei.ravel())
            lista.append(spear)
        elif dist=='ssim':
            mean_ssim = ssim(frame_ref, framei,
                              win_size=7,
                              dynamic_range=frame_ref.max() - frame_ref.min(),
                              gaussian_weights=True,
                              sigma=1.5,
                              use_sample_covariance=True)
            lista.append(mean_ssim)
        else:
            raise ValueError('Distance not recognized')
    lista = np.array(lista)

    median_cor = np.median(lista)
    mean_cor = np.mean(lista)
    if plot:
        _, ax = plt.subplots(figsize=(12,6))
        x = range(len(lista))
        ax.plot(x, lista, '-', color='blue', alpha=0.3)
        ax.plot(x, lista, '.', color='blue', alpha=0.5)
        ax.hlines(median_cor, xmin=-1, xmax=n+1, colors='purple',
                   linestyles='solid', label='Median value : '+str(median_cor))
        ax.hlines(mean_cor, xmin=-1, xmax=n+1, colors='red',
                   linestyles='solid', label='Mean value : '+str(mean_cor))
        plt.xlabel('Frame number')
        if dist=='sad':
            plt.ylabel('SAD - Manhattan distance')
        elif dist=='euclidean':
            plt.ylabel('Euclidean distance')
        elif dist=='pearson':
            plt.ylabel('Pearson correlation coefficient')
        elif dist=='spearman':
            plt.ylabel('Spearman correlation coefficient')
        elif dist=='mse':
            plt.ylabel('Mean squared error')
        elif dist=='ssim':
            plt.ylabel('Structural Similarity Index')

        plt.xlim(xmin=-1, xmax=n+1)
        plt.minorticks_on()
        plt.legend(fancybox=True, framealpha=0.5, fontsize=12, loc='best')
        plt.grid(which='both')

    return lista



#! /usr/bin/env python

"""
Module with cosmetics procedures. Contains the function for bad pixel fixing. 
Also functions for cropping cubes. 
"""

from __future__ import division
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez, V. Christiaens @ UChile/ULg'
__all__ = ['cube_crop_frames',
           'cube_drop_frames',
           'frame_crop',
           'cube_correct_nan',
           'approx_stellar_position']


import numpy as np
from astropy.stats import sigma_clipped_stats
from ..stats import sigma_filter
from ..var import frame_center, get_square


def cube_crop_frames(array, size, xy=None, force=False, verbose=True,
                     full_output=False):
    """Crops frames in a cube (3d or 4d array).
    
    Parameters
    ----------
    array : array_like 
        Input 3d or 4d array.
    size : int
        Size of the desired central sub-array in each frame.
    xy : tuple of ints
        X, Y coordinates of new frame center. If you are getting the
        coordinates from ds9 subtract 1, python has 0-based indexing.
    force : bool, optional
        ``Size`` and the size of the frames must be both even or odd. With
        ``force`` set to True this condition can be avoided.
    verbose : bool optional
        If True message of completion is showed.
    full_output: bool optional
        If true, returns cenx and ceny in addition to array_view

    Returns
    -------
    array_view : array_like
        Cube with cropped frames.
        
    """
    if not (array.ndim == 3 or array.ndim == 4):
        raise TypeError('Array is not a cube, 3d or 4d array')
    if not isinstance(size, int):
        raise TypeError('Size must be integer')

    if not force:
        if array.shape[2] % 2 == 0:    # assuming square frames
            if size % 2 != 0:
                size += 1
                print('`Size` is odd (while frame size is even)')
                print('Setting `size` to {} pixels'.format(size))
        else:
            if size % 2 == 0:
                size += 1
                print('`Size` is even (while frame size is odd)')
                print('Setting `size` to {} pixels'.format(size))
    else:
        if array.shape[2] % 2 == 0: # assuming square frames, both 3d or 4d case
            if size % 2 != 0:
                msg = "Warning: the new size is odd and the original frame "
                msg += " size is even. Make sure you are setting properly `xy`"
                print(msg)
        else:
            if size % 2 == 0:
                msg = "Warning: the new size is even and the original frame "
                msg += " size is odd. Make sure you are setting properly `xy`"
                print(msg)

    if xy is not None:
        if not (isinstance(xy[0], int) or isinstance(xy[1], int)):
            raise TypeError('XY must be a tuple of integers')
    if size >= array.shape[2]:  # assuming square frames, both 3d or 4d case
        msg = "The new size is equal to or bigger than the initial frame size"
        raise ValueError(msg)
    
    # wing is added to the sides of the subframe center
    if size % 2 != 0:
        wing = int(np.floor(size / 2.))
    else:
        wing = (size / 2.) - 0.5

    if array.ndim == 3:
        if xy is not None:
            cenx, ceny = xy
        else:
            ceny, cenx = frame_center(array[0], verbose=False)

        # +1 because python doesn't include the endpoint when slicing
        array_view = array[:, int(ceny-wing):int(ceny+wing+1),
                           int(cenx-wing):int(cenx+wing+1)]
        if verbose:
            msg = "\nCube cropped; new size [{:},{:},{:}] centered at ({:},{:})"
            print(msg.format(array_view.shape[0], array_view.shape[1],
                             array_view.shape[2], cenx, ceny))

    elif array.ndim == 4:
        if xy is not None:
            cenx, ceny = xy
        else:
            ceny, cenx = frame_center(array[0, 0], verbose=False)

        # +1 because python doesn't include the endpoint when slicing
        array_view = array[:, :, int(ceny - wing):int(ceny + wing + 1),
                           int(cenx - wing):int(cenx + wing + 1)]
        if verbose:
            msg = "\nCube cropped; new size [{:},{:},{:},{:}] "
            msg += "centered at ({:},{:})"
            print(msg.format(array_view.shape[0], array_view.shape[1],
                             array_view.shape[2], array_view.shape[3],
                             cenx, ceny))

    # Option to return the cenx and ceny
    if full_output:
        return array_view, cenx, ceny
    else:
        return array_view


def frame_crop(array, size, cenxy=None, force=False, verbose=True):
    """ Crops a square frame (2d array). Uses the ``get_square`` function.

    Parameters
    ----------
    array : array_like
        Input frame.
    size : int, odd
        Size of the subframe.
    cenxy : tuple, optional
        Coordinates of the center of the subframe.
    force : bool, optional
        Size and the size of the 2d array must be both even or odd. With
        ``force`` set to True this condition can be avoided.
    verbose : bool optional
        If True message of completion is showed.
        
    Returns
    -------
    array_view : array_like
        Sub array.
        
    """
    if not array.ndim == 2:
        raise TypeError('Array is not a frame or 2d array')
    if size >= array.shape[0]:
        msg = 'Cropping size is equal or larger than the original size'
        raise RuntimeError(msg)
    
    if not cenxy:
        ceny, cenx = frame_center(array, verbose=False)
    else:
        cenx, ceny = cenxy
    array_view = get_square(array, size, ceny, cenx, force=force)
    
    if verbose:
        print("\nDone frame cropping")
    return array_view


def cube_drop_frames(array, n, m):
    """Discards frames at the beginning or the end of a cube (axis 0).

    Parameters
    ----------
    array : array_like
        Input 3d array, cube.
    n : int
        Index of the first frame to be kept. Frames before this one are dropped.
    m : int
        Index of the last frame to be kept. Frames after this one are dropped.

    Returns
    -------
    array_view : array_like
        Cube with new size (axis 0).

    """
    if m > array.shape[0]:
        raise TypeError('End index must be smaller than the # of frames')

    array_view = array[n:m+1, :, :].copy()
    print("Done discarding frames from cube")
    return array_view


def frame_remove_stripes(array):
    """ Removes unwanted stripe artifact in frames with non-perfect bias or sky
    subtraction. Encountered this case on an LBT data cube.
    """
    lines = array[:50]
    lines = np.vstack((lines, array[-50:]))
    mean = lines.mean(axis=0)
    for i in range(array.shape[1]):
        array[:,i] = array[:,i] - mean[i]
    return array


def cube_correct_nan(cube, neighbor_box=3, min_neighbors=3, verbose=False,
                     half_res_y=False):
    """Sigma filtering of nan pixels in a whole frame or cube. Tested on
    SINFONI data.

    Parameters
    ----------
    cube : cube_like
        Input 3d or 2d array.
    neighbor_box : int, optional
        The side of the square window around each pixel where the sigma and
        median are calculated for the nan pixel correction.
    min_neighbors : int, optional
        Minimum number of good neighboring pixels to be able to correct the
        bad/nan pixels.
    verbose: bool, optional
        Whether to print more information or not during processing
    half_res_y: bool, optional
        Whether the input data have every couple of 2 rows identical, i.e. there
        is twice less angular resolution vertically than horizontally (e.g.
        SINFONI data). The algorithm goes twice faster if this option is
        rightfully set to True.

    Returns
    -------
    obj_tmp : array_like
        Output cube with corrected nan pixels in each frame
    """
    def nan_corr_2d(obj_tmp):
        n_x = obj_tmp.shape[1]
        n_y = obj_tmp.shape[0]

        if half_res_y:
            if n_y % 2 != 0:
                msg = 'The input frames do not have an even number of rows. '
                msg2 = 'Hence, you should probably not be using the option '
                msg3 = 'half_res_y = True.'
                raise ValueError(msg + msg2 + msg3)
            n_y = int(n_y / 2)
            frame = obj_tmp
            obj_tmp = np.zeros([n_y, n_x])
            for yy in range(n_y):
                obj_tmp[yy] = frame[2 * yy]

        # tuple with the 2D indices of each nan value of the frame
        nan_indices = np.where(np.isnan(obj_tmp))
        nan_map = np.zeros_like(obj_tmp)
        nan_map[nan_indices] = 1
        nnanpix = int(np.sum(nan_map))
        # Correct nan with iterative sigma filter
        obj_tmp = sigma_filter(obj_tmp, nan_map, neighbor_box=neighbor_box,
                               min_neighbors=min_neighbors, verbose=verbose)
        if half_res_y:
            frame = obj_tmp
            n_y = 2 * n_y
            obj_tmp = np.zeros([n_y, n_x])
            for yy in range(n_y):
                obj_tmp[yy] = frame[int(yy / 2)]

        return obj_tmp, nnanpix
    ############################################################################

    obj_tmp = cube.copy()

    ndims = obj_tmp.ndim
    if ndims != 2 and ndims != 3:
        raise TypeError("Input object is not two or three dimensional")

    if neighbor_box < 3 or neighbor_box % 2 == 0:
        raise ValueError('neighbor_box should be an odd value greater than 3')
    max_neigh = sum(range(3, neighbor_box + 2, 2))
    if min_neighbors > max_neigh:
        min_neighbors = max_neigh
        msg = "Warning! min_neighbors was reduced to " + str(max_neigh) + \
              " to avoid bugs. \n"
        print(msg)

    if ndims == 2:
        obj_tmp, nnanpix = nan_corr_2d(obj_tmp)
        if verbose:
            print('\n There were ', nnanpix, ' nan pixels corrected.')

    elif ndims == 3:
        n_z = obj_tmp.shape[0]
        for zz in range(n_z):
            obj_tmp[zz], nnanpix = nan_corr_2d(obj_tmp[zz])
            if verbose:
                msg = 'In channel ' + str(zz) + ', there were ' + str(nnanpix)
                msg2 = ' nan pixels corrected.'
                print(msg + msg2)

    if verbose:
        print('All nan pixels are corrected.')

    return obj_tmp


def approx_stellar_position(cube, fwhm, return_test=False, verbose=False):
    """FIND THE APPROX COORDS OF THE STAR IN EACH CHANNEL (even the ones
    dominated by noise)

    Parameters
    ----------
    obj_tmp : array_like
        Input 3d cube
    fwhm : float or array 1D
        Input full width half maximum value of the PSF for each channel.
        This will be used as the standard deviation for Gaussian kernel
        of the Gaussian filtering.
        If float, it is assumed the same for all channels.
    return_test: bool, optional
        Whether the test result vector (a bool vector with whether the star
        centroid could be find in the corresponding channel) should be returned
        as well, along with the approx stellar coordinates.
    verbose: bool, optional
        Chooses whether to print some additional information.

    Returns:
    --------
    Array of y and x approx coordinates of the star in each channel of the cube
    if return_test: it also returns the test result vector
    """
    from ..phot import peak_coordinates

    obj_tmp = cube.copy()
    n_z = obj_tmp.shape[0]

    if isinstance(fwhm, float) or isinstance(fwhm, int):
        fwhm_scal = fwhm
        fwhm = np.zeros((n_z))
        fwhm[:] = fwhm_scal

    # 1/ Write a 2-columns array with indices of all max pixel values in the cube
    star_tmp_idx = np.zeros([n_z, 2])
    star_approx_idx = np.zeros([n_z, 2])
    test_result = np.ones(n_z)
    for zz in range(n_z):
        star_tmp_idx[zz] = peak_coordinates(obj_tmp[zz], fwhm[zz])

    # 2/ Detect the outliers in each column
    _, med_y, stddev_y = sigma_clipped_stats(star_tmp_idx[:, 0], sigma=2.5)
    _, med_x, stddev_x = sigma_clipped_stats(star_tmp_idx[:, 1], sigma=2.5)
    lim_inf_y = med_y - 3 * stddev_y
    lim_sup_y = med_y + 3 * stddev_y
    lim_inf_x = med_x - 3 * stddev_x
    lim_sup_x = med_x + 3 * stddev_x

    if verbose:
        print("median y of star - 3sigma = ", lim_inf_y)
        print("median y of star + 3sigma = ", lim_sup_y)
        print("median x of star - 3sigma = ", lim_inf_x)
        print("median x of star + 3sigma = ", lim_sup_x)

    for zz in range(n_z):
        if ((star_tmp_idx[zz, 0] < lim_inf_y) or (
                star_tmp_idx[zz, 0] > lim_sup_y) or
                (star_tmp_idx[zz, 1] < lim_inf_x) or (
                        star_tmp_idx[zz, 1] > lim_sup_x)):
            test_result[zz] = 0

    # 3/ Replace by the median of neighbouring good coordinates if need be
    for zz in range(n_z):
        if test_result[zz] == 0:
            ii = 1
            inf_neigh = max(0, zz - ii)
            sup_neigh = min(n_z - 1, zz + ii)
            while test_result[inf_neigh] == 0 and test_result[sup_neigh] == 0:
                ii = ii + 1
                inf_neigh = max(0, zz - ii)
                sup_neigh = min(n_z - 1, zz + ii)
            if test_result[inf_neigh] == 1 and test_result[sup_neigh] == 1:
                star_approx_idx[zz] = np.floor((star_tmp_idx[sup_neigh] + \
                                                star_tmp_idx[inf_neigh]) / 2.)
            elif test_result[inf_neigh] == 1:
                star_approx_idx[zz] = star_tmp_idx[inf_neigh]
            else:
                star_approx_idx[zz] = star_tmp_idx[sup_neigh]
        else:
            star_approx_idx[zz] = star_tmp_idx[zz]

    if return_test:
        return star_approx_idx, test_result.astype(bool)
    else:
        return star_approx_idx



#! /usr/bin/env python

"""
Functions for timing other functions/procedures.
"""
from __future__ import print_function
__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['time_ini', 'timing', 'time_fin']

from datetime import datetime
from .utils_conf import sep

def time_ini(verbose=True):
    """Sets and prints the time in which the script started.
    
    Returns
    -------
    start_time : string
        Starting time.
    """
    start_time = datetime.now()
    if verbose:
        print(sep)
        print("Starting time: " + start_time.strftime("%Y-%m-%d %H:%M:%S"))
        print(sep)
    return start_time


def timing(start_time):
    """Prints the execution time of a script. It requires the initialization 
    with the function time_ini().
    """
    print("Running time:  " + str(datetime.now()-start_time))
    print(sep)


def time_fin(start_time):
    """Returns the execution time of a script. It requires the initialization 
    with the function time_ini().
    """
    return str(datetime.now()-start_time)

   



"""
Subpackage ``conf`` contains configuration functions and internal utilities. It
also contains a dictionary with parameters of high-contrast imaging instruments.\
"""
from __future__ import absolute_import

from .mem import *
from .param import *
from .timing import *
from .utils_conf import *
